def add_positions(pos1, pos2):
    """
    Get a new Position where two positions are added.

    NOTE: Position is a two element tuple: (x, y)

    :param Position pos1
    :param Position pos2
    :return added Position
    :rtype Position
    """

    return (pos1[0] + pos2[0], pos1[1] + pos2[1])


def offset_rect(offset, rect):
    """
    Get a new Rect where rect is translated by offset.

    NOTE: Rect is a tuple having following structure: ((left, top), (right, bottom))

    :param Position offset
    :param Rect rect
    :return translated Rect
    :rtype Rect
    """

    top_left, bottom_right = rect
    return add_positions(offset, top_left), add_positions(offset, bottom_right)


def bounding_box(rects):
    """
    Get a new Rect which circumscribes all the rects.

    :param list rects
    :return bounding Rect
    :rtype Rect
    """

    top = float('inf')
    left = float('inf')
    bottom = 0
    right = 0
    for rect in rects:
        top_left, bottom_right = rect
        top = min(top, top_left[1])
        left = min(left, top_left[0])
        bottom = max(bottom, bottom_right[1])
        right = max(right, bottom_right[0])

    return ((left, top), (right, bottom))


def padding_box(rect, padding):
    """
    Get a new Rect where rect is enlarged by padding width.

    :param Rect rect
    :param int padding
    :return enlarged Rect
    :rtype Rect
    """

    top_left, bottom_right = rect

    return (add_positions(top_left, (-padding, -padding)),
            add_positions(bottom_right, (padding, padding)))


def bounding_boxes_by_line(rects):
    """
    Get bounding Rects by line from list of Rects.

    :param list rects
    :return generator of bouding Rects
    :rtype generator
    """

    if not rects:
        return

    oneline_rects = [rects.pop(0)]
    while rects:
        rect = rects.pop(0)
        x = rect[0][0]
        prev_rect = oneline_rects[-1]
        prev_x = prev_rect[0][0]
        if prev_x <= x:
            oneline_rects.append(rect)
        else:
            # Detect line break
            yield bounding_box(oneline_rects)
            oneline_rects = [rect]

    yield bounding_box(oneline_rects)



from pyocr.tesseract import CharBoxBuilder
from pyocr.builders import Box


class ModifiedCharBoxBuilder(CharBoxBuilder):

    def __init__(self, image_height):
        """
        param: int image_height
        """

        super().__init__()

        # Though CharBoxBuilder's tesseract_configs includes 'batch.nochop',
        # this cause misrecognition. So it is removed.
        self.tesseract_configs = ['-psm', '6', 'makebox']

        self.image_height = image_height

    def read_file(self, file_descriptor):
        boxes = CharBoxBuilder.read_file(file_descriptor)

        # Though CharBoxBuilder's base position (0, 0) is top left,
        # box file's base position (0, 0) is bottom left.
        # Therefore position must be reflected.
        # See: https://github.com/tesseract-ocr/tesseract/wiki/TrainingTesseract
        for box in boxes:
            box.position = self._reflect_rect_vertically(box.position)

        def _ensure_one_char_per_box(boxes):
            """
            Occasionally, a Box contains two characters.
            So, ensure that all the boxes contains only one characters.
            """

            char_boxes = []
            for box in boxes:
                if len(box.content) == 1:
                    char_boxes.append(box)
                else:
                    assert len(box.content) >= 2
                    for c in box.content:
                        char_boxes.append(Box(c, box.position))

            return char_boxes

        return _ensure_one_char_per_box(boxes)

    def _reflect_rect_vertically(self, rect):
        """
        Get a new Rect where rect is reflected vertically.

        param: Rect rect
        return: vertically reflected Rect
        rtype: Rect
        """

        top_left, bottom_right = rect
        bottom_left = self._reflect_pos_vertically(top_left)
        top_right = self._reflect_pos_vertically(bottom_right)

        left, bottom = bottom_left
        right, top = top_right

        return ((left, top), (right, bottom))

    def _reflect_pos_vertically(self, pos):
        """
        Get a new Position where pos is reflected vertically.

        param: Position pos
        return: vertically reflected Position
        rtype: Position
        """

        return pos[0], self.image_height - pos[1]



import re

from setuptools import setup


with open('masecret/__init__.py') as f:
    init_content = f.read()

version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                    init_content, re.MULTILINE).group(1)

long_description = """
masecret
========

masecret is a command to mask secret information in image files using OCR.

Prerequisite
------------

- Python 3.3+
- `Tesseract <https://github.com/tesseract-ocr/tesseract>`__

  - Language data for OCR (can be specified with ``--lang``, default is ``eng``)
    must be available.

Installation
------------

::

    $ pip3 install masecret

You may need ``sudo``.

masecret depends on `pyocr <https://github.com/jflesch/pyocr>`__ and
`Pillow <https://pillow.readthedocs.io/>`__. If you fail to install
Pillow, please see `the installation instruction of Pillow <http://pillow.readthedocs.io/en/latest/installation.html>`__.

Usage
-----

Mask a single image file with a regular expression pattern that match AWS account number::

    $ masecret -r '[-\d]{12,}' original.png -o masked.png

Mask multiple image files (output directory must exist)::

    $ masecret -r '[-\d]{12,}' original1.png original2.png ... -o masked_images/

Mask image files in-place with ``-i`` option::

    $ masecret -i -r '[-\d]{12,}' original1.png original2.png ...

WARNING: No backup files will be saved.

SECRETS.txt
~~~~~~~~~~~

If ``-r`` option is not specified, regular expression will be read from a file named
``SECRETS.txt`` in a current directory.
Content of the file is regular expression patterns that match secret information
you want to mask. You can include multiple patterns line by line.

Full Usage
~~~~~~~~~~

::

    usage:
        masecret [options] INPUT -o OUTPUT
        masecret [options] INPUT... -o OUTPUT
        masecret -i [options] INPUT...

    Mask secret information in image files using OCR. Put regular expression
    matches secret information into a file named SECRETS.txt or -r option.

    positional arguments:
      INPUT                 input files

    optional arguments:
      -h, --help            show this help message and exit
      -V, --version         show program's version number and exit
      -o OUTPUT, --output OUTPUT
                            output file or directory (default: None)
      -r REGEX, --regex REGEX
                            regular expression matches secret information
                            (default: None)
      -s SECRET_PATH, --secret SECRET_PATH
                            path to file containing regexes line by line that
                            match secret information (default: ./SECRETS.txt)
      -l LANG, --lang LANG  language for OCR, can be multiple languages joined by
                            + sign, e.g. eng+jpn (default: eng)
      -c COLOR, --color COLOR
                            color to fill secrets (default: #666)
      -i, --in-place        mask image files in-place. WARNING: No backup files
                            will be saved (default: False)
      --tesseract-params PARAMS
                            (Advanced Option) additional parameters passed to
                            tesseract (default: -psm 6 makebox)

Debug
-----

If images are not masked as expected, the environment variable ``DEBUG``
will help you. If ``DEBUG`` is set, all the characters tesseract
recognized are printed with position.

::

    $ DEBUG=1 masecret original.png -o masked.png
    Processing original.png...
    . ((136, 90), (160, 114))
    . ((176, 90), (200, 114))
    . ((216, 90), (240, 114))
    I ((292, 104), (304, 126))
    I ((308, 104), (320, 126))
    A ((326, 104), (340, 120))
    W ((341, 104), (361, 120))
    S ((362, 103), (375, 120))
    M ((385, 104), (401, 120))
    a ((404, 108), (415, 120))
    n ((417, 108), (427, 120))
    a ((430, 108), (440, 120))
    g ((443, 108), (453, 125))
    e ((456, 108), (467, 120))
    m ((469, 108), (485, 120))
    e ((488, 108), (499, 120))
    n ((501, 108), (511, 120))
    t ((513, 105), (519, 120))
    C ((528, 103), (542, 120))
    o ((545, 108), (556, 120))
    n ((559, 108), (569, 120))
    ...

License
-------

MIT License. See: ``LICENSE``.
"""

setup(
    name="masecret",
    version=version,
    author="orangain",
    author_email="orangain@gmail.com",
    description="A command to mask secret information of images using OCR",
    long_description=long_description,
    license="MIT",
    keywords="ocr image screenshot secret mask",
    url="https://github.com/orangain/masecret",
    packages=['masecret'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
    ],
    install_requires=['Pillow', 'pyocr>=0.4.3'],
    entry_points={
        'console_scripts': [
            'masecret = masecret.cli:main'
        ]
    }
)



import sys
import os
import re
import argparse
import shlex

from PIL import Image, ImageDraw, ImageColor
from pyocr.tesseract import image_to_string

from masecret import __version__
from masecret.builders import ModifiedCharBoxBuilder
from masecret.position_utils import offset_rect, padding_box, bounding_boxes_by_line


parser = argparse.ArgumentParser(
    usage='''
    %(prog)s [options] INPUT -o OUTPUT
    %(prog)s [options] INPUT... -o OUTPUT
    %(prog)s -i [options] INPUT...''',
    description='''
        Mask secret information in image files using OCR.
        Put regular expression matches secret information
        into a file named SECRETS.txt or -r option.''',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('input_paths', metavar='INPUT', nargs='+',
                    help='input files')
parser.add_argument('-V', '--version', action='version',
                    version='%(prog)s {0}'.format(__version__))
parser.add_argument('-o', '--output', dest='output_location', metavar='OUTPUT',
                    help='output file or directory')
parser.add_argument('-r', '--regex', dest='regex', default=None,
                    help='regular expression matches secret information')
parser.add_argument('-s', '--secret', dest='secret_path', default='./SECRETS.txt',
                    help='path to file containing regexes line by line that match secret information')
parser.add_argument('-l', '--lang', dest='lang', default='eng',
                    help='language for OCR, can be multiple languages joined by + sign, e.g. eng+jpn')
parser.add_argument('-c', '--color', dest='color', default='#666',
                    help='color to fill secrets')
parser.add_argument('-i', '--in-place', dest='in_place', action='store_true', default=False,
                    help='mask image files in-place. WARNING: No backup files will be saved')
parser.add_argument('--tesseract-params', dest='tesseract_params', metavar='PARAMS',
                    default=' '.join(ModifiedCharBoxBuilder(0).tesseract_configs),
                    help='(Advanced Option) additional parameters passed to tesseract')


def main():
    args = parse_args()

    secret_res = get_secret_res(args)
    options = {
        'lang': args.lang,
        'fill_color': ImageColor.getrgb(args.color),
        'tesseract_configs': shlex.split(args.tesseract_params),
    }

    for input_path, output_path in input_output_pairs(args):
        mask_secrets(input_path, output_path, secret_res, **options)


def parse_args(args=None):
    """
    Parse command line arguments and convert to a Namespace object.

    param: list args
    return: parsed arguments
    rtype: Namespace
    """

    args = parser.parse_args(args)

    if args.in_place and args.output_location:
        parser.error('You MUST NOT specify both -i and -o options.')

    if not args.in_place:
        if not args.output_location:
            parser.error('-o OUTPUT is required unless -i is specified.')

        if len(args.input_paths) >= 2 and not os.path.isdir(args.output_location):
            parser.error('OUTPUT must be a directory when there are multiple INPUTs.')

    if args.regex and args.secret_path != './SECRETS.txt':
        parser.error('You MUST NOT specify both -r and -s options.')

    return args


def get_secret_res(args):
    """
    Get secret regexes from a Namespace object.

    param: Namespace args
    return: list of regexes
    rtype: list
    """

    if args.regex:
        return [re.compile(args.regex)]
    else:
        return read_secret_res_from_file(args.secret_path)


def read_secret_res_from_file(secret_path):
    """
    Read secret regexes from a file secret_path.

    param: str secret_path
    return: list of regexes
    rtype: list
    """

    res = []
    with open(secret_path) as f:
        for line in f:
            pattern = line.rstrip()
            res.append(re.compile(pattern))

    return res


def input_output_pairs(args):
    """
    Yield pairs of input path and output path.

    param: Namespace args
    return: generator of tuple (input_path, output_path)
    rtype: generator
    """

    for input_path in args.input_paths:
        if args.in_place:
            yield (input_path, input_path)
        elif os.path.isdir(args.output_location):
            yield (input_path, os.path.join(args.output_location, os.path.basename(input_path)))
        else:
            yield (input_path, args.output_location)


def mask_secrets(input_path, output_path, secret_res, lang, fill_color, tesseract_configs=None):
    """
    Mask secret infomation in an image.

    param: str input_path
    param: str output_path
    param: list secret_res
    param: str lang
    param: str tesseract_configs
    param: tuple fill_color
    """

    print('Processing {0}...'.format(input_path), file=sys.stderr)

    image = Image.open(input_path)
    secret_rects = find_secret_rects(image, secret_res, lang, tesseract_configs)
    print('Found {0} secrets at {1}'.format(len(secret_rects), secret_rects), file=sys.stderr)
    for rect in secret_rects:
        mask_rect(image, rect, fill_color)

    image.save(output_path)
    print('Saved to {0}'.format(output_path), file=sys.stderr)


def find_secret_rects(image, secret_res, lang, tesseract_configs=None):
    """
    Find secret rects in an image.

    param: Image image
    param: list secret_res
    param: str lang
    param: str tesseract_configs
    return: list of rects
    rtype: list
    """

    # When using pyocr, an input image is converted to RGB (not RGBA).
    # During the conversion, transparent pixels are converted into BLACK.
    # Sometimes the black pixels get in the way of recognizing text.
    #
    # For example, macOS's screenshot image taken by Command+Shift+4+Space
    # has transparent pixels around an window. This results in a black and
    # thick border in the edge of image. The border worsen quality of OCR
    # of text near by the border. To avoid this, convert transparent pixels
    # into WHITE by pasting image into an white background.
    #
    # See: http://stackoverflow.com/questions/9166400/convert-rgba-png-to-rgb-with-pil
    if image.mode == 'RGBA':
        background = Image.new('RGB', image.size, (255, 255, 255))
        background.paste(image, mask=image.split()[3])  # Paste image masked by alpha channel [3]
        image = background

    # offset = (0, 150)
    # cropped_image = image.crop((offset[0], offset[1], image.size[0], 220))
    offset = (0, 0)
    cropped_image = image

    builder = ModifiedCharBoxBuilder(cropped_image.size[1])
    if tesseract_configs:
        builder.tesseract_configs = tesseract_configs

    boxes = image_to_string(cropped_image, lang=lang, builder=builder)

    if os.environ.get('DEBUG'):
        for box in boxes:
            print(box.content, box.position)

    content = ''.join(box.content for box in boxes)
    assert len(boxes) == len(content)

    secret_rects = []
    for secret_re in secret_res:
        for m in secret_re.finditer(content):
            matched_boxes = boxes[m.start():m.end()]
            matched_rects = [b.position for b in matched_boxes]
            for rect in bounding_boxes_by_line(matched_rects):
                rect = offset_rect(offset, padding_box(rect, 2))
                secret_rects.append(rect)

    return secret_rects


def mask_rect(image, rect, color):
    """
    Fill a rect in an image.

    param: Image image
    param: Rect rect
    param: tuple color
    """

    draw = ImageDraw.Draw(image)
    draw.rectangle(rect, fill=color)


if __name__ == '__main__':
    main()
